"""
eox_tagging URL Configuration
"""
from django.urls import include, re_path

from eox_tagging import views
from eox_tagging.api_schema import docs_ui_view

urlpatterns = [
    re_path(r'^eox-info$', views.info_view, name='eox-info'),
    re_path(r'api/', include('eox_tagging.api.urls')),
    re_path(r'^api-docs/$', docs_ui_view, name='apidocs-ui'),
]



"""
Viewset for Tags.
"""
from edx_api_doc_tools import query_parameter, schema_for
from rest_framework import status, viewsets
from rest_framework.authentication import SessionAuthentication

from eox_tagging.api.v1.filters import FilterBackend, TagFilter
from eox_tagging.api.v1.pagination import TagApiPagination
from eox_tagging.api.v1.permissions import EoxTaggingAPIPermission
from eox_tagging.api.v1.serializers import TagSerializer
from eox_tagging.edxapp_accessors import get_site
from eox_tagging.edxapp_wrappers.bearer_authentication import BearerAuthentication
from eox_tagging.models import Tag

try:
    from eox_audit_model.decorators import audit_method
except ImportError:
    def audit_method(*args, **kwargs):  # pylint: disable=unused-argument
        """Identity decorator"""
        return lambda x: x


@schema_for(
    "create",
    """
    Creates a tag for a given object
    There are three different types of objects that can be labeled with a tag: courses,\
    users and enrollments. The type of objects that can be labeled (and extra\
    validations for different fields) are defined in the configuration of the\
    site.

    **Example Request**

        POST /eox-tagging/api/v1/tags/
        {
            "tag_type": "subscription_level",
            "tag_value": "premium",
            "target_type": "courseoverview",
            "target_id": "course-v1:edX+DemoX+Demo_Course",
            "access": "PUBLIC",
            "owner_type": "site"
        }

    **Parameters**

    - `tag_type` (**required**, string, _body_):
    General category for the tag (i.e subscription_level). This value is set in\
    the site configuration.

    - `tag_value` (**required**, string, _body_):
    An item of the category (i.e premium). If there isn't a validation in the\
    site configuration it can take any string.

    - `target_type` (**required**, string, _body_):
    One of courseoverview, user, courseenrollment, generatedcertificate

    - `target_id` (**required**, string, _body_): Identifier of the target\
      object. For users, username; for courseoverview, course_id and for\
      courseenrollments/generatedcertificates a string with the following format: "`username:\
      course_id`. For generatedcertificates it can also be its verify_uuid"

    - `activation_date` (**optional**, string, _body_):
    DateTime format `YYYY-MM-DD HH:MM:SS`.

    - `expiration_date` (**optional**, string, _body_):
    DateTime format `YYYY-MM-DD HH:MM:SS`.

    - `owner_type` (**optional**, string, _body_):
    Owner of the tag, either `site` or `user`

    - `access` (**optional**, string, _body_):
    Visibility of the tag, either `PUBLIC`, `PRIVATE` or `PROTECTED`
    """,
)
@schema_for(
    "destroy",
    """
    Delete single tag by key. Deleted tags are inactivated (soft delete)
    """,
    responses={status.HTTP_404_NOT_FOUND: "Not found"},
)
@schema_for(
    "retrieve",
    """
    Fetch details for a single tag by key
    """,
    responses={status.HTTP_404_NOT_FOUND: "Not found"},
)
@schema_for(
    "list",
    """
    Fetch a list of tags.

    The list can be narrowed using the available filters.
    - Some filters are incompatible with each other,\
    namely `course_id`, `username` and `target_type`. The reason being that `course_id` and `username`\
    have an implicit `target_type` of `courseoverview` and `user`.
    - DateTime filters must have the following format `YY-MM-DD HH:MM:SS`. Time is optional, date is not.\
    Time must be UTC.
    - Parameters not defined bellow will be ignored. If you apply a filter with a typo you'll get the \
    whole list of tags.
    """,
    parameters=[
        query_parameter(
            "key",
            str,
            "The unique identifier. Same as `GET /eox-tagging/api/v1/tags/{key}`",
        ),
        query_parameter(
            "status", str, "Filter active or inactive tags. Default: active"
        ),
        query_parameter(
            "include_inactive", bool, "If true include the inactive tags on the list. Default false"
        ),
        query_parameter(
            "tag_type",
            str,
            "The type of the tag, set on the configuration of the site (i.e. Subscription level)",
        ),
        query_parameter("tag_value", str, "The value of the tag (i.e. Premium)"),
        query_parameter(
            "course_id",
            str,
            "Shortcut to filter objects of target_type `courseoverview` with id `course_id`.",
        ),
        query_parameter(
            "username",
            str,
            "Shortcut to filter objects of target_type `user` with id `username`.",
        ),
        query_parameter(
            "certificate_verify_uuid",
            str,
            "Shortcut to filter objects of target_type `generatedcertificate` with id `verify_uuid`.",
        ),
        query_parameter(
            "owner_type",
            str,
            "Shortcut to filter objects of owner_type `user` or `site`.",
        ),
        query_parameter(
            "target_type",
            str,
            "The type of the object that was tagged, one of: `course (use opaquekeyproxymodel for`"
            "course overview types), `courseenrollment`, `user`, `generatedcertificate`",
        ),
        query_parameter(
            "enrollment_username",
            str,
            "User identifier (username) to be used when target_type=courseenrollment. "
            "Can be omitted and is ignored for a different target_type",
        ),
        query_parameter(
            "enrollment_course_id",
            str,
            "Course identifier to be used when target_type=courseenrollment."
            "Can be omitted and is ignored for a different target_type",
        ),
        query_parameter(
            "certificate_username",
            str,
            "User identifier (username) to be used when target_type=generatedcertificate. "
            "Can be omitted and is ignored for a different target_type",
        ),
        query_parameter(
            "certificate_course_id",
            str,
            "Course identifier to be used when target_type=generatedcertificate."
            "Can be omitted and is ignored for a different target_type",
        ),
        query_parameter(
            "created_at_before",
            str,
            "Filter tags created before date. Format `YY-MM-DD HH:MM:SS`",
        ),
        query_parameter(
            "created_at_after",
            str,
            "Filter tags created after date. Format `YY-MM-DD HH:MM:SS`",
        ),
        query_parameter(
            "activation_date_before",
            str,
            "Filter tags created before date. Format `YY-MM-DD HH:MM:SS`",
        ),
        query_parameter(
            "activation_date_after",
            str,
            "Filter tags created after date. Format `YY-MM-DD HH:MM:SS`",
        ),
        query_parameter(
            "expiration_date_before",
            str,
            "Filter tags created before date. Format `YY-MM-DD HH:MM:SS`",
        ),
        query_parameter(
            "expiration_date_after",
            str,
            "Filter tags created after date. Format `YY-MM-DD HH:MM:SS`",
        ),
        query_parameter("access", str, "Filter by access, One of `PUBLIC`, `PRIVATE`, `PROTECTED`"),
    ],
    responses={status.HTTP_404_NOT_FOUND: "Not found"},
)
class TagViewSet(viewsets.ModelViewSet):  # pylint: disable=too-many-ancestors
    """Viewset for listing and creating Tags."""

    serializer_class = TagSerializer
    authentication_classes = (BearerAuthentication, SessionAuthentication)
    permission_classes = (EoxTaggingAPIPermission,)
    pagination_class = TagApiPagination
    filter_backends = (FilterBackend,)
    filterset_class = TagFilter
    lookup_field = "key"
    http_method_names = ["get", "post", "delete", "head"]

    def get_queryset(self):
        """Restricts the returned tags."""
        queryset = Tag.objects.all()

        queryset = self.__get_objects_by_status(queryset)

        queryset = self.__get_objects_by_owner(queryset)

        return queryset

    def create(self, request, *args, **kwargs):
        """Hijack the create method and use a wrapper function to perform the
        audit process. The original parameters of create are not very useful in
        raw form, this way we pass more useful information to our wrapper
        function to be audited
        """

        @audit_method(action="eox_tagging-api-v1-viewset:tagviewset-create")
        def audited_create(body):  # pylint: disable=unused-argument
            return super(TagViewSet, self).create(request, *args, **kwargs)  # pylint: disable=super-with-arguments

        return audited_create(body=request.data)

    def destroy(self, request, *args, **kwargs):
        """Hijack the destroy method and use a wrapper function to perform the
        audit process. The original parameters of destroy are not very useful in
        raw form, this way we pass more useful information to our wrapper
        function to be audited
        """

        @audit_method(action="eox_tagging-api-v1-viewset:tagviewset-destroy")
        def audited_destroy(path):  # pylint: disable=unused-argument
            return super(TagViewSet, self).destroy(request, *args, **kwargs)  # pylint: disable=super-with-arguments

        return audited_destroy(path=request.path)

    def __get_objects_by_status(self, queryset):
        """Method that returns queryset filtered by tag status."""
        include_inactive = self.request.query_params.get("include_inactive")

        if "key" in self.request.query_params or self.action == "retrieve":
            include_inactive = "true"

        if not include_inactive or include_inactive.lower() not in ["true", "1"]:
            queryset = queryset.active()

        return queryset

    def __get_objects_by_owner(self, queryset):
        """Method that returns queryset filtered by tag owner"""
        owner_type = self.request.query_params.get("owner_type")
        owner_information = self.__get_request_owner(owner_type)

        try:
            queryset_union = queryset.none()
            for owner in owner_information:
                queryset_union |= queryset.find_by_owner(**owner)

            return queryset_union
        except Exception:  # pylint: disable=broad-except
            return queryset.none()

    def __get_request_owner(self, owner_type):
        """Returns the owners of the tag to filter the queryset."""
        site = self.__get_site()
        user = self.__get_user()

        if not owner_type:
            return [site, user]

        if owner_type.lower() == "user":
            return [user]

        if owner_type.lower() == "site":
            return [site]

        return []

    def __get_site(self):
        """Returns the current site."""
        site = get_site()

        return {
            "owner_id": {"id": site.id},
            "owner_type": "site",
        }

    def __get_user(self):
        """Returns the current user."""
        user = self.request.user

        return {
            "owner_id": {"username": user.username},
            "owner_type": "user",
        }



"""
Admin class.
"""
from django.contrib import admin, messages
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponseRedirect
from opaque_keys import InvalidKeyError
from opaque_keys.edx.keys import CourseKey

from eox_tagging.forms import TagForm
from eox_tagging.models import OpaqueKeyProxyModel, Tag


class TagAdmin(admin.ModelAdmin):
    """Tag admin."""
    list_display = [
        "key",
        "tag_type",
        "tag_value",
        "tagged_object",
        "owner",
        "status",
    ]
    readonly_fields = (
        'target_as_nice_string',
        'status',
        'created_at',
        'inactivated_at',
    )
    search_fields = ('tag_type', 'tag_value', 'status')

    form = TagForm

    def target_as_nice_string(self, tag):
        """
        Render the opaque key proxy as a nice Course Key or any other target as its unicode.
        """
        return str(tag.target_object)

    def get_search_results(self, request, queryset, search_term):
        """
        Custom search to support searching on the tagged objects
        """
        queryset, use_distinct = super().get_search_results(
            request,
            queryset,
            search_term
        )
        # TODO: we need to connect the TagQuery search here.
        return queryset, use_distinct

    def owner(self, tag):
        """
        Displays useful info about the owner of the tag.
        """
        # pylint: disable=broad-except
        try:
            return f"{tag.owner_object_type}: {tag.owner_object}"
        except Exception as error:
            return str(error)

    def tagged_object(self, tag):
        """
        Displays useful info about the tagged object.
        """
        # pylint: disable=broad-except
        try:
            if tag.target_object:
                return f"{tag.target_object_type}: {tag.target_object}"
            return f"Resource locator: {tag.resource_locator}"
        except Exception as error:
            return str(error)

    def add_view(self, request, form_url='', extra_context=None):
        """
        Custom method to handle the specific case of tagging course_keys

        The POST request is modified to add the target object if and only if an opaque_key is defined
        and no target_object_id is specified.

        We decided to do this because we needed to add the target object before the Tag instance was
        created.
        """
        should_intervene = True

        if not request.POST:
            should_intervene = False

        if request.POST.get('target_type', None) and request.POST.get('target_object_id', None):
            should_intervene = False

        if not request.POST.get('opaque_key', None):
            should_intervene = False

        if should_intervene:
            try:
                course_key = CourseKey.from_string(request.POST.get('opaque_key'))
                opaque_key_proxy, _ = OpaqueKeyProxyModel.objects.get_or_create(opaque_key=course_key)
            except InvalidKeyError:
                opaque_key = request.POST['opaque_key']
                message = f"EOX_TAGGING | Error: Opaque Key {opaque_key} does not match with opaque_keys.edx definition"
                messages.error(request, message)
                return HttpResponseRedirect(request.path)

            request.POST = request.POST.copy()
            request.POST['target_type'] = ContentType.objects.get(model='OpaqueKeyProxyModel').id
            request.POST['target_object_id'] = opaque_key_proxy.id

        return super().add_view(request, form_url='', extra_context=None)


admin.site.register(Tag, TagAdmin)



"""
Authentication definitions.
"""
from importlib import import_module

from django.conf import settings


def get_bearer_authentication():
    """ Gets BearerAuthentication function. """

    backend_function = settings.EOX_TAGGING_BEARER_AUTHENTICATION
    backend = import_module(backend_function)

    return backend.get_bearer_authentication()


BearerAuthentication = get_bearer_authentication()



"""
Swagger view generator
"""
from django.conf import settings
from django.urls import include, re_path, reverse
from drf_yasg.generators import OpenAPISchemaGenerator
from drf_yasg.openapi import SwaggerDict
from drf_yasg.views import get_schema_view
from edx_api_doc_tools import get_docs_cache_timeout, internal_utils, make_api_info
from rest_framework import permissions


class APISchemaGenerator(OpenAPISchemaGenerator):
    """
    Schema generator for eox-core.

    Define specific security definition using oauth without overwritting project wide
    settings.
    """

    def get_security_definitions(self):
        security_definitions = {
            "OAuth2": {
                "flow": "application",
                "tokenUrl": f"{settings.LMS_ROOT_URL}{reverse('access_token')}",
                "type": "oauth2",
            },
        }
        security_definitions = SwaggerDict.as_odict(security_definitions)
        return security_definitions


api_urls = [
    re_path(r"eox-tagging/api/", include("eox_tagging.api.urls"))
]

api_info = make_api_info(
    title="eox tagging",
    version="v1",
    email=" contact@edunext.co",
    description=internal_utils.dedent("""\
    **eox tagging REST API**

    eox-tagging is a small plugin that provides the ability to ‘tag’ or label\
    certain objects of the Open edX platform (Courses, Course Enrollments,\
    Users). The tag contains the object that is being tagged, its identifier, a\
    ‘tag_type’ indicating the set of values the tag can have and ‘tag_value’\
    which holds a value of type ‘tag_type’. Additionally one can provide extra\
    info such as: Access level, or timestamps for when the tag should be considered\
    active. \

    eox tagging is meant to be a lightweight plugin with emphasis on flexibility,\
    most of the logic regarding the deactivation of tags at a given time must be\
    handled separately.
    """),
)

docs_ui_view = get_schema_view(
    api_info,
    generator_class=APISchemaGenerator,
    public=True,
    permission_classes=[permissions.AllowAny],
    patterns=api_urls,
).with_ui("swagger", cache_timeout=get_docs_cache_timeout())



""" Backend abstraction. """
from importlib import import_module

from django.conf import settings


def get_course_overview(*args, **kwargs):
    """ Get course overview object. """
    backend_function = settings.EOX_TAGGING_GET_COURSE_OVERVIEW
    backend = import_module(backend_function)
    return backend.get_course_overview(*args, **kwargs)


CourseOverview = get_course_overview()



"""
Integration tests.

This suite performs multiple http requests to guarantee
that the eox-tagging rest API is behaving as expected on a live server.
"""
import json
from os import environ
from unittest import skipUnless

import requests
from django.test import SimpleTestCase
from django.urls import reverse
from rest_framework import status


@skipUnless(environ.get("TEST_INTEGRATION"), "integration")
class TagTestCase(SimpleTestCase):  # pragma: no cover
    """Test suite."""

    @classmethod
    def setUpClass(cls):
        """Initialize data fixtures."""
        cls.data = json.loads(environ.get("TEST_DATA"))
        base_url = cls.data["base_url"]
        endpoint = reverse("tag-list")

        cls.url = f"{base_url}/eox-tagging{endpoint}"
        cls.session = requests.Session()
        cls.session.headers.update(
            {
                "Authorization": f"Bearer {cls.data['access_token']}",
            }
        )

        cls.tags_data = [
            {
                "tag_type": "test_user_tag",
                "tag_value": "test_user_tag",
                "target_type": "user",
                "target_id": cls.data["username"],
                "access": "PUBLIC",
                "owner_type": "site",
            },
            {
                "tag_type": "test_course_tag",
                "tag_value": "test_course_tag",
                "target_type": "courseoverview",
                "target_id": cls.data["course_id"],
                "access": "PUBLIC",
                "owner_type": "site",
            },
            {
                "tag_type": "test_enrollment_tag",
                "tag_value": "test_enrollment_tag",
                "target_type": "courseenrollment",
                "target_id": f"{cls.data['username']}: {cls.data['course_id']}",
                "access": "PUBLIC",
                "owner_type": "site",
            },
        ]

        cls.tear_down_tags_ids = []
        cls.delete_test_tag_id = create_tag(cls.session, cls.url, cls.tags_data[0])
        cls.read_test_tag_id = create_tag(cls.session, cls.url, cls.tags_data[0])
        cls.tear_down_tags_ids.append(cls.read_test_tag_id)

    @classmethod
    def tearDownClass(cls):
        """Deactive tags created by tests."""
        for tag_id in cls.tear_down_tags_ids:
            delete_tag(cls.session, cls.url, tag_id)

    def test_read_list_tag(self):
        """Test the tag list view."""
        response = self.session.get(self.url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("count", response.json())

    def test_read_detail_tag(self):
        """Test the tag detail view."""
        tag_id = self.read_test_tag_id
        url = f"{self.url}{tag_id}/"
        expected_response = {
            "key": tag_id,
            "tag_value": self.tags_data[0]["tag_value"],
            "tag_type": self.tags_data[0]["tag_type"],
            "access": "PUBLIC",
            "activation_date": None,
            "expiration_date": None,
            "status": "ACTIVE",
        }

        response = self.session.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictContainsSubset(expected_response, response.json())

    def test_create_tag_user(self):
        """Test tag creation for target=user."""
        response = self.session.post(self.url, self.tags_data[0])

        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.text)

        self.tear_down_tags_ids.append(response.json()["key"])

    def test_create_tag_course(self):
        """Test tag creation for target=course."""
        response = self.session.post(self.url, self.tags_data[1])

        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.text)

        self.tear_down_tags_ids.append(response.json()["key"])

    def test_create_tag_enrollment(self):
        """Test tag creation for target=enrollment."""
        response = self.session.post(self.url, self.tags_data[2])

        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.text)

        self.tear_down_tags_ids.append(response.json()["key"])

    def test_create_tag_fail_validation(self):
        """Test creation denial due to validation."""
        data = self.tags_data[0].copy()
        data["access"] = "PRIVATE"

        response = self.session.post(self.url, data)

        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.text
        )

    def test_delete_tag(self):
        """Test tag deletion."""
        tag_id = self.delete_test_tag_id
        url = f"{self.url}{tag_id}/"

        response = self.session.delete(url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)


def create_tag(session, url, tag_data):
    """Auxiliary function to create tags."""
    response = session.post(url, data=tag_data)
    response.raise_for_status()
    return response.json()["key"]


def delete_tag(session, url, tag_id):
    """Auxiliary function to delete tags."""
    url = f"{url}{tag_id}/"
    response = session.delete(url)
    response.raise_for_status()



""" File to define validations for tag model fields.
"""
import datetime
import logging
import re

import crum
import opaque_keys.edx.keys as all_opaque_keys
import six
from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from eox_core.edxapp_wrapper.certificates import get_generated_certificate
from eox_core.edxapp_wrapper.enrollments import get_enrollment
from eox_core.edxapp_wrapper.users import get_edxapp_user
from opaque_keys import InvalidKeyError  # pylint: disable=ungrouped-imports, useless-suppression

from eox_tagging.edxapp_wrappers.course_overview import CourseOverview

log = logging.getLogger(__name__)

DATETIME_FORMAT_VALIDATION = "%Y-%m-%d %H:%M:%S"
GeneratedCertificate = get_generated_certificate()


class TagValidators:
    """Defines all validator methods."""

    def __init__(self, instance):
        """
        Attributes:
            instance: instance of the model to validate before saving
            definitions: configuration matching fields to validate
        """
        self.instance = instance
        self.model_validations = {
            "User": self.__validate_user_integrity,
            "OpaqueKeyProxyModel": self.__validate_proxy_model,
            "CourseEnrollment": self.__validate_enrollment_integrity,
            "Site": self.__validate_site_integrity,
            "GeneratedCertificate": self.__validate_certificate_integrity,
        }
        self.__configuration_types = {
            "in": list,
            "equals": six.string_types,
            "regex": six.string_types,
            "exist": bool,
            "opaque_key": six.string_types,
        }
        self.current_tag_definitions = {}
        self.set_configuration()

    def set_configuration(self):
        """Function that sets and validates configuration for model instance."""
        self.validate_no_updating()  # Don't validate if trying to update
        self.__select_configuration()
        self.__force_configuration_values()
        self.__validate_configuration_types()
        self.__validate_configuration()

    def __select_configuration(self):
        """Function that selects the correct configuration for model instance."""
        for tag_def in settings.EOX_TAGGING_DEFINITIONS:
            tag_type = tag_def.get("tag_type")
            if tag_type == self.instance.tag_type:
                self.current_tag_definitions = tag_def
                return
        raise ValidationError(f"Tag_type '{self.instance.tag_type}' not configured")

    def __force_configuration_values(self):
        """
        Function that sets tag values defined in the configuration.

        For example:
        {
            "force_access": "public"
        }
        Then the access level of the tag must be set tu public, no matter if it had a value before.
        Also, using force the validations are skipped.
        """
        pattern = "force_"
        configurations_copy = dict(self.current_tag_definitions)

        for key, value in configurations_copy.items():
            if key.startswith(pattern):
                try:
                    self.instance.set_attribute(key.replace(pattern, ""), value)
                except Exception:
                    raise ValidationError(f"EOX_TAGGING | The field {key} with value `{value}` is wrongly configured")
                del self.current_tag_definitions[key]

    def __validate_required(self):
        """Function that validates the configuration for the required fields target and owner."""
        required_target_fields = r".*target_object|.*resource_locator"
        required_owner_fields = r".*owner_object"

        if not self.__find_attribute(required_target_fields):
            raise ValidationError(f"The target object for `tag_type`: '{self.instance.tag_type}' is not configured.")

        if not self.__find_attribute(required_owner_fields):
            self.current_tag_definitions["validate_owner_object"] = "site"

    def __find_attribute(self, attr_pattern):
        """Function that fiends an attribute in the current tag definitions."""
        for attr in self.current_tag_definitions:
            if re.match(attr_pattern, attr):
                return True
        return False

    # Config validations
    def __validate_configuration(self):
        """
        Function that validates EOX_TAGGING_DEFINITIONS. The validations consist in:
            - Validate available validations
            - Validate field names
            - Check required fields in configuration
        If any error occur a ValidationError will be raised.
        """
        self.__validate_required()  # Validate that the require configs exists
        regex = r"validate_"

        for key, value in self.current_tag_definitions.items():

            # Validate value correctness if it has validations defined
            if re.match(regex, key) and not isinstance(value, six.string_types):
                for _key in value:  # Validations must exist as a class method
                    try:
                        getattr(self, f"validate_{_key}")
                    except AttributeError:
                        raise ValidationError(
                            f"EOX_TAGGING | The field {key} with value `{_key}` is wrongly configured."
                        )
            # Validate key existence
            clean_key = re.sub(regex, "", key)
            try:
                self.instance.get_attribute(clean_key)
            except AttributeError:
                raise ValidationError(f"EOX_TAGGING | The field `{key}` is wrongly configured.")

    def __validate_configuration_types(self):
        """Function that validate the correct type for pairs <key, value> in configuration."""

        regex = r"validate_"

        for key, value in self.current_tag_definitions.items():

            if re.match(regex, key) and isinstance(value, dict):

                for key_ in value:

                    value_type = self.__configuration_types.get(key_)
                    field_value = value.get(key_)

                    if value_type and not isinstance(field_value, value_type):
                        raise ValidationError(
                            f"EOX_TAGGING | The validation '{key_}' for '{key}' is wrongly configured."
                        )

            elif not isinstance(value, six.string_types):
                raise ValidationError(f"EOX_TAGGING | The field '{key}' is wrongly configured.")

    # GFK validations

    def validate_fields_integrity(self):
        """Helper function that calls for every object that needs integrity validation."""
        fields_to_validate = ["owner_object", "target_object"]
        map(self.__validate_model, fields_to_validate)

    def __validate_model(self, field_name):
        """Function that validates the instances in GFK fields calling the integrity validators."""
        try:
            model_name = self.instance.get_attribute(field_name, name=True)
        except AttributeError:
            raise ValidationError(f"EOX_TAGGING | The field '{field_name}' is wrongly configured.")
        try:
            if model_name:
                self.model_validations[model_name](field_name)
        except KeyError:
            raise ValidationError(f"EOX_TAGGING | Could not find integrity validation for field '{field_name}'")

    # Integrity validators
    def __validate_proxy_model(self, object_name):
        """
        Function that validates existence of proxy model.

        Arguments:
            - object_name: name of the object to validate. It can be: target_object or owner_object
        """
        opaque_key = self.instance.get_attribute(object_name).opaque_key
        try:
            CourseOverview.get_from_id(opaque_key)
        except Exception:
            raise ValidationError(
                f"EOX_TAGGING | Could not find opaque key: '{opaque_key}' for relation '{object_name}'"
            )

    def __validate_user_integrity(self, object_name):
        """
        Function that validates existence of user.

        Arguments:
            - object_name: name of the object to validate. It can be: target_object or owner_object
        """
        request = crum.get_current_request()
        data = {
            "username": self.instance.get_attribute(object_name).username,  # User needs to have username
            "site": request.site,
        }
        try:
            get_edxapp_user(**data)

        except Exception:
            raise ValidationError(f"EOX_TAGGING | Could not find ID: {self.instance.get_attribute(object_name).id} \
                for relation '{object_name}'")

    def __validate_enrollment_integrity(self, object_name):
        """
        Function that validates existence of the enrollment.

        Arguments:
            - object_name: name of the object to validate. It can be: target_object or owner_object
        """
        object_ = self.instance.get_attribute(object_name)
        data = {
            "username": object_.username,
            "course_id": str(object_.course_id),
        }
        try:
            enrollment, _ = get_enrollment(**data)
            if not enrollment:
                raise ValidationError(
                    f"EOX_TAGGING | Enrollment for user '{data['username']}' and courseID '{data['course_id']}' \
                        does not exist"
                )
        except Exception:
            raise ValidationError(
                f"EOX_TAGGING | Error getting enrollment for user '{data['username']}' \
                    and courseID '{data['course_id']}'"
            )

    def __validate_site_integrity(self, object_name):
        """
        Function that validates existence of the site.

        Arguments:
            - object_name: name of the object to validate. It can be: target_object or owner_object
        """
        site_id = self.instance.get_attribute(object_name).id

        try:
            Site.objects.get(id=site_id)
        except ObjectDoesNotExist:
            raise ValidationError(f"EOX_TAGGING | Site '{site_id}' does not exist")

    def __validate_certificate_integrity(self, object_name):
        """
        Function that validates existence of the certificate object.

        Arguments:
            - object_name: name of the object to validate.
        """
        certificate_id = self.instance.get_attribute(object_name).id

        try:
            GeneratedCertificate.objects.get(id=certificate_id)
        except ObjectDoesNotExist:
            raise ValidationError(f"EOX_TAGGING | Certificate '{certificate_id}' does not exist")

    # Other validations
    def validate_no_updating(self):
        """Function that validates that the save is not an update."""
        if self.instance.id:
            #  Exception raised when trying to update
            raise ValidationError("EOX_TAGGING | Can't update tag. Tags are immutable by definition")

    def validate_fields(self):
        """Function that validates all fields for the current definition."""

        regex = r"^validate_"

        for key, value in self.current_tag_definitions.items():

            clean_key = re.sub(regex, "", key)
            if isinstance(value, six.string_types):
                validator_method = getattr(self, "validate_equals")
                validator_method(clean_key, value)
            else:
                for _key, _value in value.items():
                    validator_method = getattr(self, f"validate_{_key}")
                    validator_method(clean_key, _value)

    def validate_opaque_key(self, field, value):
        """
        Function that if called validates that that field is value OpaqueKey defined in
        opaque_keys.edx.keys.

        Arguments:
            - field: field to validate
            - value: validations defined for the field
        """
        field_value = self.instance.get_attribute(field)
        try:
            opaque_key_to_validate = getattr(all_opaque_keys, value)
            # Validation method for OpaqueKey: opaque_key_to_validate
            getattr(opaque_key_to_validate, "from_string")(field_value)
        except InvalidKeyError:
            # We don't recognize this key
            raise ValidationError(f"The key '{field_value}' for '{field}' is not an opaque key")

    def validate_in(self, field, values):
        """
        Function that validates that the field exists in values.

        Arguments:
            - field: field to validate
            - value: validations defined for the field
        """
        field_value = self.instance.get_attribute(field, name=True)

        if isinstance(field_value, datetime.datetime):
            field_value = str(field_value)

        formatted_values = [item.lower() for item in values]
        formatted_field_value = field_value.lower()

        if formatted_field_value not in formatted_values:
            # Values allowed is list of values (at least one)

            raise ValidationError(f"EOX_TAGGING | The field '{field}' is not in tag definitions.")

    def validate_exist(self, field, value):  # pylint: disable=unused-argument
        """
        Function that validates that the field exists, this means, is not None.

        Arguments:
            - field: field to validate
            - value: validations defined for the field
        """
        field_value = self.instance.get_attribute(field)

        if not field_value:
            raise ValidationError(f"EOX_TAGGING | The field '{field}' is required.")

    def validate_equals(self, field, value):
        """
        Function that validates that the field_value is equal to value.

        Arguments:
            - field: field to validate
            - value: validations defined for the field
        """
        field_value = self.instance.get_attribute(field, name=True)

        if isinstance(field_value, datetime.datetime):
            self.__compare_equal_dates(field_value, value)
            return

        if not field_value:
            raise ValidationError(f"EOX_TAGGING | The field '{field}' is required and must be equal to '{value}'.")
        if field_value.lower() != value.lower():
            raise ValidationError(f"EOX_TAGGING | The field '{field}' must be equal to '{value}'.")

    def validate_between(self, field, value):
        """
        Function that checks that a date must be in between two dates.

        Arguments:
            - field: field to validate
            - value: validations defined for the field
        """
        field_value = self.instance.get_attribute(field)
        datetime_obj = []

        for datetime_str in value:
            try:
                datetime_obj.append(datetime.datetime.strptime(datetime_str, DATETIME_FORMAT_VALIDATION))
            except TypeError:
                raise ValidationError(
                    f"EOX_TAGGING | The DateTime field '{datetime_str}' \
                        must follow the format '{DATETIME_FORMAT_VALIDATION}'."
                )

        if field_value < datetime_obj[0] or field_value > datetime_obj[-1]:
            raise ValidationError(
                f"EOX_TAGGING | The DateTime field '{field_value}' \
                    must be in between '{str(datetime_obj[0])}' and '{str(datetime_obj[-1])}."
            )

    def __compare_equal_dates(self, field_value, value):
        """
        Function that checks that a date must be equal to another date.

        Arguments:
            - field_value: datetime to validate.
            - value: datetime strings to validate against.
        """
        try:
            datetime_str = datetime.datetime.strptime(value, DATETIME_FORMAT_VALIDATION)
        except TypeError:
            raise ValidationError(
                f"EOX_TAGGING | The DateTime field '{value}' must follow the format '{DATETIME_FORMAT_VALIDATION}'."
            )

        if field_value != datetime_str:
            raise ValidationError(f"EOX_TAGGING | The DateTime field '{field_value}' must be equal to '{str(value)}'.")

    def validate_regex(self, field, value):
        """
        Function that validates that the field matches value.
        Arguments:
            - field: field to validate
            - value: validations defined for the field
        """
        field_value = self.instance.get_attribute(field, name=True)

        if not re.search(value, field_value):
            # Values allowed is regex pattern
            raise ValidationError(f"EOX_TAGGING | The field '{field}' is not in tag definitions.")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-06-18 19:35
from __future__ import unicode_literals

import uuid

import django.db.models.deletion
import opaque_keys.edx.django.models
from django.db import migrations, models

import eox_tagging.constants


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='OpaqueKeyProxyModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('opaque_key', opaque_keys.edx.django.models.CourseKeyField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.UUIDField(default=uuid.uuid4, editable=False, unique=True, verbose_name='Public identifier')),
                ('tag_value', models.CharField(max_length=150)),
                ('tag_type', models.CharField(max_length=150)),
                ('access', models.PositiveIntegerField(choices=[(1, 'PUBLIC'), (2, 'PROTECTED'), (3, 'PRIVATE')], default=eox_tagging.constants.AccessLevel(1))),
                ('activation_date', models.DateField(blank=True, null=True)),
                ('expiration_date', models.DateField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('status', models.PositiveIntegerField(choices=[(0, 'INVALID'), (1, 'VALID')], default=eox_tagging.constants.Status(1), editable=False)),
                ('invalidated_at', models.DateTimeField(blank=True, editable=False, null=True)),
                ('target_object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('owner_object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('owner_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='owner_tag_type', to='contenttypes.ContentType')),
                ('target_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tag_type', to='contenttypes.ContentType')),
            ],
            options={
                'verbose_name': 'tag',
                'verbose_name_plural': 'tags',
            },
        ),
    ]



"""
Router module for Tag API.
"""
from rest_framework.routers import DefaultRouter

from eox_tagging.api.v1.viewset import TagViewSet

router = DefaultRouter()
router.register(r'tags', TagViewSet, basename='tag')



"""
Production Django settings for eox_tagging project.
"""
from __future__ import unicode_literals


def plugin_settings(settings):  # pylint: disable=unused-argument
    """
    Set of plugin settings used by the Open Edx platform.
    More info: https://github.com/edx/edx-platform/blob/master/openedx/core/djangoapps/plugins/README.rst
    """



"""
File for constans definitions.
"""
from enum import IntEnum


class AccessLevel(IntEnum):
    """
    Class that defines access for tags
    The more higher the class, the most restrictions has.
    """
    PUBLIC = 1
    PROTECTED = 2
    PRIVATE = 3

    @classmethod
    def choices(cls):
        """Returns choices for the class"""
        return [(key.value, key.name) for key in cls]  # pylint: disable=not-an-iterable, useless-suppression

    @classmethod
    def get_choice(cls, value):
        """Function that gets choice value for AccessLevel"""
        for key in cls:  # pylint: disable=not-an-iterable, useless-suppression
            if key.value == value:
                return key.name
        return None

    @classmethod
    def get_access_object(cls, value):
        """Function that given a key name returns an access object."""
        for key in cls:  # pylint: disable=not-an-iterable, useless-suppression
            if key.name.lower() == value.lower():
                return key
        return None


class Status(IntEnum):
    """
    Class that defines status for tags.
    When a tag is created, is created with a valid tag, when someone soft deletes
    the tag, then is marked as invalid.
    """
    ACTIVE = 1
    INACTIVE = 0

    @classmethod
    def choices(cls):
        """Returns choices for the class."""
        return [(key.value, key.name) for key in cls]  # pylint: disable=not-an-iterable, useless-suppression



""" Test classes for Tags viewset. """
import datetime
from functools import wraps

from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.test import TestCase, override_settings
from django.urls import reverse
from mock import patch
from rest_framework.test import APIClient

from eox_tagging.api.v1.serializers import TagSerializer
from eox_tagging.constants import AccessLevel
from eox_tagging.models import Tag


def mock_decorator(action=None):  # pylint: disable=unused-argument
    """Utility decorator that returns same function."""
    def decorator(func):
        @wraps(func)
        def decorated_function(*args, **kwargs):
            return func(*args, **kwargs)
        return decorated_function
    return decorator


@override_settings(
    EOX_TAGGING_DEFINITIONS=[
        {
            "tag_type": "example_tag_1",
            "validate_owner_object": "User",  # default = Site
            "validate_access": {"equals": "PRIVATE"},
            "validate_tag_value": {"in": ["example_tag_value", "example_tag_value_1"]},
            "validate_target_object": "User",
            "validate_expiration_date": {"exist": True},
        },
        {
            "tag_type": "example_tag_2",
            "validate_owner_object": "Site",
            "validate_access": {"equals": "PRIVATE"},
            "validate_tag_value": {"in": ["example_tag_value", "example_tag_value_1"]},
            "validate_target_object": "User",
            "validate_expiration_date": {"exist": True},
        },
        {
            "tag_type": "example_tag_3",
            "validate_owner_object": "User",
            "validate_access": {"equals": "PUBLIC"},
            "validate_tag_value": {"in": ["example_tag_value", "example_tag_value_1"]},
            "validate_target_object": "Site",
            "validate_expiration_date": {"exist": True},
        },
    ])
class TestTagViewSet(TestCase):  # pylint: disable=too-many-instance-attributes
    """Test class for tags viewset."""

    patch_permissions = patch(
        "eox_tagging.api.v1.permissions.EoxTaggingAPIPermission.has_permission",
        return_value=True,
    )
    patch("eox_tagging.api.v1.viewset.audit_method", mock_decorator).start()

    def setUp(self):
        """Model setup used to create objects used in tests."""
        self.target_object = User.objects.create(username="user_test")
        self.owner_site = Site.objects.get(id=settings.TEST_SITE)

        # Client authentication
        self.owner_user = User.objects.create(username="myuser")
        self.client = APIClient()
        self.client.force_authenticate(self.owner_user)

        self.example_tag = Tag.objects.create(
            tag_value="example_tag_value",
            tag_type="example_tag_1",
            target_object=self.target_object,
            owner_object=self.owner_user,
            access=AccessLevel.PRIVATE,
            expiration_date=datetime.datetime(2020, 10, 19, 10, 20, 30),
            activation_date=datetime.datetime(2020, 10, 19, 10, 20, 30),
        )

        self.example_tag_1 = Tag.objects.create(
            tag_value="example_tag_value",
            tag_type="example_tag_2",
            target_object=self.target_object,
            owner_object=self.owner_site,
            access=AccessLevel.PRIVATE,
            expiration_date=datetime.datetime(2020, 9, 19, 10, 30, 40),
            activation_date=datetime.datetime(2020, 9, 19, 10, 30, 40),
        )

        self.example_tag_2 = Tag.objects.create(
            tag_value="example_tag_value",
            tag_type="example_tag_2",
            target_object=self.target_object,
            owner_object=self.owner_site,
            access=AccessLevel.PRIVATE,
            expiration_date=datetime.datetime(2020, 8, 19, 10, 30, 40),
        )

        self.example_tag_3 = Tag.objects.create(
            tag_value="example_tag_value",
            tag_type="example_tag_3",
            target_object=self.owner_site,
            owner_object=self.owner_user,
            access=AccessLevel.PUBLIC,
            expiration_date=datetime.datetime(2020, 7, 19, 10, 30, 40),
        )
        self.key = self.example_tag.key.hex
        # Test URLs
        self.url = reverse("tag-list")
        self.url_details = reverse("tag-detail", args=[self.key])

    @patch_permissions
    def test_get_all_tags(self, _):
        """Used to test getting all tags."""
        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 200)

    @patch_permissions
    def test_retreive_tag(self, _):
        """Used to test getting a tag given its key."""
        response = self.client.get(self.url_details)

        self.assertEqual(response.status_code, 200)

    @patch_permissions
    def test_create_tag(self, _):
        """"Used to test creating a tag."""

        data = {
            "tag_type": "example_tag_1",
            "tag_value": "example_tag_value",
            "target_type": "user",
            "target_id": "user_test",
            "owner_type": "user",
            "access": "PRIVATE",
            "expiration_date": "2020-12-04 10:20:30",
        }

        response = self.client.post(self.url, data, format='json')

        self.assertEqual(response.status_code, 201)

    @patch_permissions
    def test_create_tag_without_owner(self, _):
        """"
        Used to test creating a tag without an owner. The owner should be set as a default
        to be the site.
        """
        data = {
            "tag_type": "example_tag_2",
            "tag_value": "example_tag_value",
            "target_type": "user",
            "target_id": "user_test",
            "access": "PRIVATE",
            "expiration_date": "2020-12-04 10:20:30",
        }

        response = self.client.post(self.url, data, format='json')

        owner_type = response.data.get("meta").get("owner_type")
        self.assertEqual(owner_type, "Site")

    @patch_permissions
    def test_create_tag_with_iso_datetime_format(self, _):
        """"Used to test creating a tag using ISO format in datetime fields."""
        data = {
            "tag_type": "example_tag_1",
            "tag_value": "example_tag_value",
            "target_type": "user",
            "target_id": "user_test",
            "owner_type": "user",
            "access": "PRIVATE",
            "expiration_date": "2020-12-04T10:20:30.15785",
        }

        response = self.client.post(self.url, data, format='json')

        self.assertEqual(response.status_code, 201)

    @patch_permissions
    def test_create_tag_with_wrong_datetime_format(self, _):
        """"
        Used to test creating a tag using wrong format in datetime fields. This results in a
        bad request.
        """
        data = {
            "tag_type": "example_tag_1",
            "tag_value": "example_tag_value",
            "target_type": "user",
            "target_id": "user_test",
            "owner_type": "user",
            "access": "PRIVATE",
            "expiration_date": "12-04-2020 10:20:30",
        }

        response = self.client.post(self.url, data, format='json')

        self.assertEqual(response.status_code, 400)

    @patch_permissions
    def test_create_tag_with_owner_site(self, _):
        """"Used to test creating a tag with site as owner."""
        data = {
            "tag_type": "example_tag_2",
            "tag_value": "example_tag_value",
            "target_type": "user",
            "target_id": "user_test",
            "owner_type": "site",
            "access": "PRIVATE",
            "expiration_date": "2020-12-04 10:30:40",
        }

        response = self.client.post(self.url, data, format='json')

        owner_type = response.data.get("meta").get("owner_type").lower()
        self.assertEqual(owner_type, "site")

    @patch_permissions
    def test_create_tag_with_wrong_owner(self, _):
        """"Used to test creating a tag with wrong owner_type. This results in bad request."""
        data = {
            "tag_type": "example_tag_1",
            "tag_value": "example_tag_value",
            "target_type": "user",
            "target_id": "user_test",
            "owner_type": "course",  # default is site
            "access": "PRIVATE",
            "expiration_date": "2020-12-04 10:30:40 ",
        }

        response = self.client.post(self.url, data, format='json')

        self.assertEqual(response.status_code, 400)

    @patch_permissions
    def test_patch_tag(self, _):
        """Used to test that a tag can't be updated."""
        response = self.client.patch(self.url_details)

        self.assertEqual(response.status_code, 405)

    @patch_permissions
    def test_put_tag(self, _):
        """Used to test that a tag can't be updated."""
        response = self.client.put(self.url_details)

        self.assertEqual(response.status_code, 405)

    @patch_permissions
    def test_filter_by_tag_key(self, _):
        """Used to test getting a tag given its key."""
        query_params = {
            "key": self.example_tag.key.hex,
        }

        response = self.client.get(self.url, query_params)

        data = response.json().get("results")[0]
        self.assertEqual(data.get("key").replace("-", ""), self.key)

    @patch_permissions
    def test_filter_by_username(self, _):
        """Used to test getting a tag given its target."""
        query_params = {
            "username": "user_test",
        }

        response = self.client.get(self.url, query_params)

        data = response.json().get("results")[0].get("meta")
        self.assertEqual(data.get("target_id"), "user_test")

    @patch_permissions
    def test_filter_by_owner_user(self, _):
        """Used to test getting a tag given its owner of type user."""
        query_params = {
            "owner_type": "user",
        }

        response = self.client.get(self.url, query_params)

        results = response.json().get("results")
        owner_is_user = [tag.get("meta").get("owner_type") == "User" for tag in results]
        self.assertTrue(all(owner_is_user))

    @patch_permissions
    def test_filter_by_owner_user_and_target(self, _):
        """Used to test getting a tag given its owner of type user and target type."""
        query_params = {
            "owner_type": "user",
            "target_type": "user",
        }

        response = self.client.get(self.url, query_params)

        results = response.json().get("results")
        owner_is_user = [tag.get("meta").get("owner_type") == "User" for tag in results]
        target_is_user = [tag.get("meta").get("target_type") == "User" for tag in results]
        self.assertTrue(all(owner_is_user))
        self.assertTrue(all(target_is_user))

    @patch_permissions
    def test_filter_by_owner_user_and_username(self, _):
        """Used to test getting a tag given its owner of type user and target username."""
        query_params = {
            "owner_type": "user",
            "username": "user_test",
        }

        response = self.client.get(self.url, query_params)

        results = response.json().get("results")
        owner_type = results[0].get("meta").get("owner_type")
        username = results[0].get("meta").get("target_id")
        self.assertEqual(owner_type, "User")
        self.assertEqual(username, "user_test")

    @patch_permissions
    def test_filter_by_owner_site(self, _):
        """Used to test getting a tag given its owner of type user."""
        query_params = {
            "owner_type": "site",
        }

        response = self.client.get(self.url, query_params)

        data = response.json().get("results")[0]
        owner_type = data.get("meta").get("owner_type").lower()
        self.assertEqual(owner_type, "site")

    @patch_permissions
    def test_filter_by_wrong_owner(self, _):
        """
        Used to test getting a tag given an undefined type of owner. This returns an empty
        queryset.
        """
        query_params = {
            "owner_type": "course",
        }

        response = self.client.get(self.url, query_params)

        data = response.json().get("results")
        self.assertFalse(data)

    @patch_permissions
    def test_filter_by_type(self, _):
        """Used to test getting a tag given its target."""
        query_params = {
            "target_type": "user",
        }

        response = self.client.get(self.url, query_params)

        self.assertEqual(response.status_code, 200)

    @patch_permissions
    def test_filter_by_existent_access(self, _):
        """Used to test filters with valid access level."""
        query_params = {
            "access": "public",
        }

        response = self.client.get(self.url, query_params)

        data = response.json().get("results")[0]
        self.assertEqual(data.get("key").replace("-", ""), self.example_tag_3.key.hex)

    @patch_permissions
    def test_filter_by_non_existent_access(self, _):
        """Used to test filters with invalid access level."""
        query_params = {
            "access": "pub",
        }

        response = self.client.get(self.url, query_params)

        data = response.json().get("results")
        self.assertFalse(data)

    @patch_permissions
    def test_filter_using_after_datetime(self, _):
        """
        Used to test filtering tags using a range datetime filter. In this case, the filter is
        before a specific datetime.
        """
        query_params = {
            "activation_date_after": "2020-10-10 10:20:30",
            "activation_date_0": "2020-10-10 10:20:30",
        }
        datetime_value = datetime.datetime(2020, 10, 10, 10, 20, 30)
        datetime_format = "%Y-%m-%dT%H:%M:%SZ"

        response = self.client.get(self.url, query_params)

        results = response.json().get("results")
        datetimes = [datetime.datetime.strptime(tag.get("activation_date"), datetime_format) > datetime_value
                     for tag in results]
        self.assertTrue(all(datetimes))

    @patch_permissions
    def test_activation_using_before_filter(self, _):
        """
        Used to test filtering tags using a range datetime filter on activation date. In this case, the filter is
        after a specific datetime.
        """
        query_params = {
            "activation_date_before": "2020-10-10 10:20:30",
            "activation_date_1": "2020-10-10 10:20:30",
        }
        datetime_value = datetime.datetime(2020, 10, 10, 10, 20, 30)
        datetime_format = "%Y-%m-%dT%H:%M:%SZ"

        response = self.client.get(self.url, query_params)

        results = response.json().get("results")
        activation_date = [datetime.datetime.strptime(tag.get("activation_date"), datetime_format) < datetime_value
                           for tag in results]
        self.assertTrue(all(activation_date))

    @patch_permissions
    def test_expiration_using_before_filter(self, _):
        """
        Used to test filtering tags using a range datetime filter on expiration date. In this case, the filter is
        after a specific datetime.
        """
        query_params = {
            "expiration_date_before": "2020-10-10 10:20:30",
            "expiration_date_1": "2020-10-10 10:20:30",
        }
        datetime_value = datetime.datetime(2020, 10, 10, 10, 20, 30)
        datetime_format = "%Y-%m-%dT%H:%M:%SZ"

        response = self.client.get(self.url, query_params)

        results = response.json().get("results")
        expiration_date = [datetime.datetime.strptime(tag.get("expiration_date"), datetime_format) < datetime_value
                           for tag in results]
        self.assertTrue(all(expiration_date))

    @patch_permissions
    def test_soft_delete(self, _):
        """Used to test a tag soft deletion."""
        response = self.client.delete(self.url_details)

        self.assertEqual(response.status_code, 204)

    @patch_permissions
    def test_getting_meta_field(self, _):
        """Used to test getting tag most important technical information."""
        response = self.client.get(self.url_details)

        data = response.json().get("meta")
        self.assertIsNotNone(data)
        self.assertEqual(data.get("target_id"), str(self.example_tag.target_object))
        self.assertEqual(data.get("target_type"), self.example_tag.target_object_type)
        self.assertEqual(data.get("owner_id"), str(self.example_tag.owner_object))
        self.assertEqual(data.get("owner_type"), self.example_tag.target_object_type)

    @patch_permissions
    def test_retreive_inactive_tag(self, _):
        """Used to test getting a tag given its key."""
        self.client.delete(self.url_details)
        response = self.client.get(self.url_details)

        self.assertEqual(response.data.get("key").replace("-", ""), self.key)
        self.assertEqual(response.data.get("status"), "INACTIVE")

    @patch_permissions
    def test_get_inactive_tag_with_filter(self, _):
        """Used to test getting a tag given its key."""
        query_params = {
            "key": self.key,
        }

        self.client.delete(self.url_details)
        response = self.client.get(self.url, query_params)

        data = response.json().get("results")[0]
        self.assertEqual(data.get("status"), "INACTIVE")
        self.assertEqual(data.get("key").replace("-", ""), self.key)

    @patch_permissions
    def test_listing_inactive_tags(self, _):
        """Used to test adding inactive tags to the list of tags."""
        query_params = {
            "include_inactive": "true"
        }

        self.client.delete(self.url_details)  # To deactivate the tag
        response_include_inactive = self.client.get(self.url, query_params)

        include_inactive = [tag.get("key") for tag in response_include_inactive.json().get("results")]
        serialized_tag = TagSerializer(self.example_tag).data
        self.assertIn(serialized_tag.get("key"), include_inactive)

    @patch_permissions
    def test_listing_just_active_tags(self, _):
        """Used to test getting just active tags using as the only inactive tag the tag in url_details."""
        self.client.delete(self.url_details)  # To deactivate the tag
        response_exclude_inactive = self.client.get(self.url)

        exclude_inactive = [tag.get("key") for tag in response_exclude_inactive.json().get("results")]
        serialized_tag = TagSerializer(self.example_tag).data
        self.assertNotIn(serialized_tag.get("key"), exclude_inactive)



"""Module that implements helper functions for other modules."""
import crum
from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from eox_core.edxapp_wrapper.certificates import get_generated_certificate
from eox_core.edxapp_wrapper.users import get_edxapp_user
from opaque_keys.edx.keys import CourseKey

from eox_tagging.edxapp_wrappers.course_overview import CourseOverview
from eox_tagging.edxapp_wrappers.enrollments import CourseEnrollment

GeneratedCertificate = get_generated_certificate()


def get_user(**kwargs):
    """Function used to get users."""
    user_id = kwargs.get("target_id")
    if getattr(settings, "EOX_TAGGING_SKIP_VALIDATIONS", False):  # Skip these validations while testing
        return User.objects.get(username=user_id)

    site = crum.get_current_request().site
    user = get_edxapp_user(username=user_id, site=site)
    return user


def get_course(**kwargs):
    """Function used to get courses from the platform."""
    course_id = kwargs.get("target_id")
    opaque_key = CourseKey.from_string(course_id)
    if getattr(settings, "EOX_TAGGING_SKIP_VALIDATIONS", False):  # Skip these validations while testing
        return object

    course = CourseOverview.get_from_id(opaque_key)
    return course


def get_site(**_kwargs):
    """Function used to get current site."""
    if getattr(settings, "EOX_TAGGING_SKIP_VALIDATIONS", False):  # Use TEST_SITE while testing
        site = Site.objects.get(id=settings.TEST_SITE)
    else:
        site = crum.get_current_request().site
    return site


def get_course_enrollment(**kwargs):
    """Function used to get enrollments from the platform."""
    username = kwargs.get("username")
    course_id = kwargs.get("course_id")

    if getattr(settings, "EOX_TAGGING_SKIP_VALIDATIONS", False):
        return object

    user = get_user(**{"target_id": username})
    course_id = CourseKey.from_string(course_id)
    return CourseEnrollment.objects.get(user_id=user.id, course_id=course_id)


def get_certificate(**kwargs):
    """
    Get GeneratedCertificate for specified id, key download URL or the course_id
    and username associated with it.
    """
    target_id = {}
    verify_uuid = kwargs.get("target_id")
    if verify_uuid:
        target_id["verify_uuid"] = kwargs.get("target_id")
    else:
        target_id["user__username"] = kwargs.get("username")
        target_id["course_id"] = CourseKey.from_string(kwargs.get("course_id"))

    return GeneratedCertificate.objects.get(**target_id)


def get_object_from_edxapp(object_type, **kwargs):
    """Helper function to get objects from edx-platfrom given its identifiers."""
    related_objects = {
        "user": get_user,
        "courseoverview": get_course,
        "site": get_site,
        "courseenrollment": get_course_enrollment,
        "generatedcertificate": get_certificate,
    }
    related_object = related_objects.get(object_type.lower())
    return related_object(**kwargs)



"""
Test classes for Tags router
"""
from django.test import TestCase
from django.urls import reverse


class TestRouters(TestCase):
    """Test class for API router."""

    def test_get_route_for_list_tags(self):
        """Used to test correctness of list route."""
        list_route = reverse("tag-list")

        self.assertEqual(list_route, "/api/v1/tags/")

    def test_get_route_for_tag_details(self):
        """Used to test correctness of details route."""
        detail_route = reverse("tag-detail", args=[1])

        self.assertEqual(detail_route, "/api/v1/tags/1/")
